#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#define CELL_SIZE 53
#define CELL_HDR  5
#define CELL_DATA (CELL_SIZE - CELL_HDR)

#define MY_VPI  8
#define MY_VCI 35

int cell_read(int fd,unsigned char *buf, int n,
			  unsigned int * vpi,
			  unsigned int * vci,
			  unsigned int * pti)
{
	unsigned char cellbuf[CELL_SIZE];
	int i;

	if (n < CELL_SIZE - CELL_HDR)
		return -1;

	if (read(fd,cellbuf,CELL_SIZE) != CELL_SIZE)
		return -1;

	memcpy(buf,cellbuf+CELL_HDR,CELL_DATA);

	*vpi = (cellbuf[0] << 4) | (cellbuf[1]>>4);
	*vci = ((cellbuf[1] & 0xf) <<12) | (cellbuf[2]<<4) | (cellbuf[3]>>4);
	*pti = (cellbuf[3] & 0xf);

	printf("VPI/VCI=%d/%d PTI=%x buf =",*vpi,*vci,*pti);
	for (i=0;i<CELL_DATA;i++)
		printf(" %02x",buf[i]);
	printf("\n");

	return CELL_DATA;
}

int cell_write(int vci,int vpi,unsigned char *buf,int n)
{
}

int aal5_read(int fd,unsigned char *buf, int n)
{
	unsigned char cell_buf[48];
	unsigned int vpi, vci, pti;

	int idx, i;
	int len = 0;

	for (idx=0;;idx++)
	{
		if (cell_read(fd,cell_buf,sizeof(cell_buf),&vpi,&vci,&pti) <0)
			return -1;

		if (vpi != MY_VPI || vci != MY_VCI)
		{
			printf("VPI/VCI mismatch ... ignored!\n");
			continue; 
		}

		memcpy(buf+idx*CELL_DATA, cell_buf, CELL_DATA);
		len += CELL_DATA;

		if (pti & 2)
			break;
	}

	printf("AAL5: ");
	for (i=0;i<len;i++)
		printf(" %02x",buf[i]);
	printf("\n");

	return len;
}

int aal5_write(int fd,unsigned char *buf, int n)
{
	// on decompose en block de 48 octets

	int nb_blocks = n / 48;
	int i;

	for (i=0;i<nb_blocks;i++)
		;
//;		cell_write(vci,vpi,buf+48*i);
}

int main(int argc,const char *argv[])
{
	int fd, out;
	int n;
	unsigned char buf[65536];

	fd = open(argv[1],O_RDONLY);
	if (fd < 0)
	{
		perror(argv[1]);
		return -1;
	}

	out = open("aal5.log",O_WRONLY|O_CREAT|O_TRUNC,0666);
	if (out < 0)
	{
		perror("creating aal5.log");
		return -1;
	}

	while ((n=aal5_read(fd,buf,sizeof(buf))) > 0)
		if (write(out,buf,n) != n)
			perror("writing aal5.log");

	close(fd);
	close (out);

	return 0;
}
