/*
  ALCATEL SpeedTouch USB modem utility : PPPoA implementation.
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : 21/03/2001

  This program is designed to work under pppd, with the option "pty".

  21/03/2001: you MUST run 'modem_init' and 'modem_link' before trying to
  run this one.

  25/03/2001: 'modem_link' has been replaced by 'modem_diag'.
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <usb.h>

/* for ident(1) command */
static char id[] = "@(#) $Id: pppoa.c,v 1.2 2001/04/12 22:06:51 papillau Exp $";

#define ST_VENDOR  0x06b9
#define ST_PRODUCT 0x4061

#define DATA_TIMEOUT 10000

#define MY_VPI  8
#define MY_VCI 35

#define CELL_SIZE 53
#define CELL_HDR  05
#define CELL_DATA 48

#define syncHDLC 1

int alternate = 2; /* seems to be the good value,
		      at least, used in speedtouch.c from Alcatel */

struct usb_dev_handle * usb_modem_init()
{
  struct usb_bus * bus;
  struct usb_device * dev, * st_dev=NULL;
  struct usb_dev_handle *st_handle;

  usb_init();
  usb_find_busses();
  usb_find_devices();

  for (bus=usb_busses;bus;bus=bus->next)
    for (dev=bus->devices;dev;dev=dev->next)
	if (dev->descriptor.idVendor == ST_VENDOR
	    && dev->descriptor.idProduct == ST_PRODUCT)
	  st_dev = dev;

  if (st_dev == NULL)
    return NULL;
     
  st_handle = usb_open(st_dev);

  if (usb_claim_interface(st_handle,1) < 0)
    {
      printf("usb_claim_interface: %s\n",usb_strerror());
      return NULL;
    }

  if (alternate!=0)
    {
      printf("setting alternate %d\n",alternate);
      if (usb_set_altinterface(st_handle,alternate) < 0)
	{
	  printf("usb_set_altinterface: %s\n",usb_strerror());
	  return NULL;
	}
  }

  /*
  if (usb_resetep(st_handle,0x07)<0)
    printf("usb_resetep 0x07: %s\n",usb_strerror());
  
  if (usb_resetep(st_handle,0x87)<0)
    printf("usb_resetep 0x87: %s\n",usb_strerror());
  */

  /*
  if (usb_clear_halt(st_handle,0x07)<0)
   printf("usb_clear_halt 0x07: %s\n",usb_strerror());
  
  if (usb_clear_halt(st_handle,0x87)<0)
    printf("usb_clear_halt 0x87: %s\n",usb_strerror());
  */

  fflush(stdout);
  return st_handle;
}

#define CRC32_REMAINDER CBF43926
#define CRC32_INITIAL 0xffffffff

#define CRC32(c,crc) (crc32tab[((size_t)(crc>>24) ^ (c)) & 0xff] ^ (((crc) << 8)))

unsigned long  crc32tab[256] =
{
 0x00000000L, 0x04C11DB7L, 0x09823B6EL, 0x0D4326D9L,
 0x130476DCL, 0x17C56B6BL, 0x1A864DB2L, 0x1E475005L,
 0x2608EDB8L, 0x22C9F00FL, 0x2F8AD6D6L, 0x2B4BCB61L,
 0x350C9B64L, 0x31CD86D3L, 0x3C8EA00AL, 0x384FBDBDL,
 0x4C11DB70L, 0x48D0C6C7L, 0x4593E01EL, 0x4152FDA9L,
 0x5F15ADACL, 0x5BD4B01BL, 0x569796C2L, 0x52568B75L,
 0x6A1936C8L, 0x6ED82B7FL, 0x639B0DA6L, 0x675A1011L,
 0x791D4014L, 0x7DDC5DA3L, 0x709F7B7AL, 0x745E66CDL,
 0x9823B6E0L, 0x9CE2AB57L, 0x91A18D8EL, 0x95609039L,
 0x8B27C03CL, 0x8FE6DD8BL, 0x82A5FB52L, 0x8664E6E5L,
 0xBE2B5B58L, 0xBAEA46EFL, 0xB7A96036L, 0xB3687D81L,
 0xAD2F2D84L, 0xA9EE3033L, 0xA4AD16EAL, 0xA06C0B5DL,
 0xD4326D90L, 0xD0F37027L, 0xDDB056FEL, 0xD9714B49L,
 0xC7361B4CL, 0xC3F706FBL, 0xCEB42022L, 0xCA753D95L,
 0xF23A8028L, 0xF6FB9D9FL, 0xFBB8BB46L, 0xFF79A6F1L,
 0xE13EF6F4L, 0xE5FFEB43L, 0xE8BCCD9AL, 0xEC7DD02DL,
 0x34867077L, 0x30476DC0L, 0x3D044B19L, 0x39C556AEL,
 0x278206ABL, 0x23431B1CL, 0x2E003DC5L, 0x2AC12072L,
 0x128E9DCFL, 0x164F8078L, 0x1B0CA6A1L, 0x1FCDBB16L,
 0x018AEB13L, 0x054BF6A4L, 0x0808D07DL, 0x0CC9CDCAL,
 0x7897AB07L, 0x7C56B6B0L, 0x71159069L, 0x75D48DDEL,
 0x6B93DDDBL, 0x6F52C06CL, 0x6211E6B5L, 0x66D0FB02L,
 0x5E9F46BFL, 0x5A5E5B08L, 0x571D7DD1L, 0x53DC6066L,
 0x4D9B3063L, 0x495A2DD4L, 0x44190B0DL, 0x40D816BAL,
 0xACA5C697L, 0xA864DB20L, 0xA527FDF9L, 0xA1E6E04EL,
 0xBFA1B04BL, 0xBB60ADFCL, 0xB6238B25L, 0xB2E29692L,
 0x8AAD2B2FL, 0x8E6C3698L, 0x832F1041L, 0x87EE0DF6L,
 0x99A95DF3L, 0x9D684044L, 0x902B669DL, 0x94EA7B2AL,
 0xE0B41DE7L, 0xE4750050L, 0xE9362689L, 0xEDF73B3EL,
 0xF3B06B3BL, 0xF771768CL, 0xFA325055L, 0xFEF34DE2L,
 0xC6BCF05FL, 0xC27DEDE8L, 0xCF3ECB31L, 0xCBFFD686L,
 0xD5B88683L, 0xD1799B34L, 0xDC3ABDEDL, 0xD8FBA05AL,
 0x690CE0EEL, 0x6DCDFD59L, 0x608EDB80L, 0x644FC637L,
 0x7A089632L, 0x7EC98B85L, 0x738AAD5CL, 0x774BB0EBL,
 0x4F040D56L, 0x4BC510E1L, 0x46863638L, 0x42472B8FL,
 0x5C007B8AL, 0x58C1663DL, 0x558240E4L, 0x51435D53L,
 0x251D3B9EL, 0x21DC2629L, 0x2C9F00F0L, 0x285E1D47L,
 0x36194D42L, 0x32D850F5L, 0x3F9B762CL, 0x3B5A6B9BL,
 0x0315D626L, 0x07D4CB91L, 0x0A97ED48L, 0x0E56F0FFL,
 0x1011A0FAL, 0x14D0BD4DL, 0x19939B94L, 0x1D528623L,
 0xF12F560EL, 0xF5EE4BB9L, 0xF8AD6D60L, 0xFC6C70D7L,
 0xE22B20D2L, 0xE6EA3D65L, 0xEBA91BBCL, 0xEF68060BL,
 0xD727BBB6L, 0xD3E6A601L, 0xDEA580D8L, 0xDA649D6FL,
 0xC423CD6AL, 0xC0E2D0DDL, 0xCDA1F604L, 0xC960EBB3L,
 0xBD3E8D7EL, 0xB9FF90C9L, 0xB4BCB610L, 0xB07DABA7L,
 0xAE3AFBA2L, 0xAAFBE615L, 0xA7B8C0CCL, 0xA379DD7BL,
 0x9B3660C6L, 0x9FF77D71L, 0x92B45BA8L, 0x9675461FL,
 0x8832161AL, 0x8CF30BADL, 0x81B02D74L, 0x857130C3L,
 0x5D8A9099L, 0x594B8D2EL, 0x5408ABF7L, 0x50C9B640L,
 0x4E8EE645L, 0x4A4FFBF2L, 0x470CDD2BL, 0x43CDC09CL,
 0x7B827D21L, 0x7F436096L, 0x7200464FL, 0x76C15BF8L,
 0x68860BFDL, 0x6C47164AL, 0x61043093L, 0x65C52D24L,
 0x119B4BE9L, 0x155A565EL, 0x18197087L, 0x1CD86D30L,
 0x029F3D35L, 0x065E2082L, 0x0B1D065BL, 0x0FDC1BECL,
 0x3793A651L, 0x3352BBE6L, 0x3E119D3FL, 0x3AD08088L,
 0x2497D08DL, 0x2056CD3AL, 0x2D15EBE3L, 0x29D4F654L,
 0xC5A92679L, 0xC1683BCEL, 0xCC2B1D17L, 0xC8EA00A0L,
 0xD6AD50A5L, 0xD26C4D12L, 0xDF2F6BCBL, 0xDBEE767CL,
 0xE3A1CBC1L, 0xE760D676L, 0xEA23F0AFL, 0xEEE2ED18L,
 0xF0A5BD1DL, 0xF464A0AAL, 0xF9278673L, 0xFDE69BC4L,
 0x89B8FD09L, 0x8D79E0BEL, 0x803AC667L, 0x84FBDBD0L,
 0x9ABC8BD5L, 0x9E7D9662L, 0x933EB0BBL, 0x97FFAD0CL,
 0xAFB010B1L, 0xAB710D06L, 0xA6322BDFL, 0xA2F33668L,
 0xBCB4666DL, 0xB8757BDAL, 0xB5365D03L, 0xB1F740B4L 
};
                                                                
unsigned int calc_crc(unsigned char *mem, int len, unsigned int initial)
{
  unsigned int crc;
  crc = initial;

  for(;len;mem++,len--)
    {
      crc = CRC32(*mem, crc);
    }
  return(crc);
}

int ppp_read(int fd,unsigned char * buf, int n)
{
  /* we must handle HDLC framing, since this is what pppd
     sends to us. I use some code from rp-pppoe
  */

  unsigned char ppp_buf[64*1024];
  int r, i;

  r = read(fd,ppp_buf,sizeof(ppp_buf));
  if (r<0)
    return r;

  /* supress leading 2 bytes */

  if (syncHDLC)
    {
      memcpy(buf,ppp_buf+2,r-2);

      printf("PPP read : len=%d",r-2);
      for (i=0;i<r-2;i++)
	printf(" %02x",buf[i]);
      printf("\n");
      fflush(stdout);

      return (r-2);
    }
  else
    {
      printf("ppp_read: async HDLC not handled\n");
    }
}

int ppp_write(int fd,unsigned char * buf, int n)
{
  unsigned char ppp_buf[64*1024];

  if (syncHDLC)
    {
      ppp_buf[0] = 0xff; //FRAME_ADDR;
      ppp_buf[1] = 0x03; //FRAME_CTRL;
      
      memcpy(ppp_buf +2, buf, n);
      return write(fd,ppp_buf,n+2);
    }
  else
    {
      printf("ppp_write: async HDLC not handled\n");
    }
}

int cell_read(struct usb_dev_handle *st_handle,
	      unsigned char *buf, int n,
	      unsigned int *vpi, unsigned int *vci,unsigned int *pti)
{
  unsigned char cellbuf[CELL_SIZE];
  int i,r;

  for (;;)
    {
      r = usb_bulk_read(st_handle,0x87,cellbuf,sizeof(cellbuf),DATA_TIMEOUT);
      if (r < 0)
	{
	  time_t t = time(NULL);
	  printf("%.24s usb_bulk_read: %s\n",ctime(&t),usb_strerror());
	  fflush(stdout);
	  if (errno != ETIMEDOUT)
	    return r;
	}
    }

  memcpy(buf,cellbuf+CELL_HDR,CELL_DATA);

  *vpi = (cellbuf[0] << 4) | (cellbuf[1]>>4);
  *vci = ((cellbuf[1] & 0xf) <<12) | (cellbuf[2]<<4) | (cellbuf[3]>>4);
  *pti = (cellbuf[3] & 0xf);
  
  printf("cell_read: VPI/VCI=%d/%d PTI=%x buf =",*vpi,*vci,*pti);
  for (i=0;i<10 /*CELL_DATA*/ ;i++)
    printf(" %02x",buf[i]);
  printf("\n");
  fflush(stdout);

  return CELL_DATA;
}

int cell_write(struct usb_dev_handle *st_handle,
	       unsigned char *buf, int n,
	       unsigned int vpi, unsigned int vci, unsigned int pti)
{
  int r,i;
  unsigned char cellbuf[CELL_SIZE];

  unsigned char gfc = 0;
  unsigned char hec = 0xec;

  if (n != CELL_DATA)
    return -1;

  /* normalize */

  vpi &= 0xff;
  vci &= 0xffff;
  pti &= 0xf;

  cellbuf[0] = (gfc<<4) | (vpi>>4);
  cellbuf[1] = (vpi<<4) | (vci>>12);
  cellbuf[2] = (vci>>4);
  cellbuf[3] = ((vci&0xf) << 4) | pti;
  cellbuf[4] = hec;

  memcpy(cellbuf+CELL_HDR,buf,CELL_DATA);

  printf("writing a cell\n");
  printf("VPI/VCI=%d/%d PTI=%x buf =",vpi,vci,pti);
  for (i=0;i<CELL_SIZE;i++)
    printf(" %02x",cellbuf[i]);
  printf("\n");
  fflush(stdout); 


  if (st_handle)
    {
      r = usb_bulk_write(st_handle,0x07,cellbuf,CELL_SIZE,DATA_TIMEOUT);
      if (r<0)
	{
	  printf("usb_bulk_write: %s\n",usb_strerror());
	  fflush(stdout);
	  return r;
	}
      
      printf("cell written: ok\n");
      fflush(stdout);
    }

  return CELL_DATA;
}

int aal5_read(struct usb_dev_handle *st_handle,
	      unsigned char * buf, int n)
{
  unsigned char cell_buf[CELL_DATA];
  unsigned int vpi, vci, pti;
  
  int idx, i;
  int len = 0;
  unsigned int real_len;
  
  for (idx=0;;idx++)
    {
      if (cell_read(st_handle,cell_buf,sizeof(cell_buf),&vpi,&vci,&pti) <0)
	return -1;
      
      if (vpi != MY_VPI || vci != MY_VCI)
	{
	  printf("VPI/VCI mismatch ... ignored!\n");
	  continue;
	}
      
      memcpy(buf+idx*CELL_DATA, cell_buf, CELL_DATA);
      len += CELL_DATA;
      
      
      if (pti & 2)
	break;
    }

  real_len = (buf[len -6] << 8) | buf[len -5];
  
  printf("AAL5: len=%d => real_len=%d",len,real_len);
  for (i=0;i<len;i++)
    printf(" %02x",buf[i]);
  printf("\n");
  
  return len;
}

int aal5_write(struct usb_dev_handle *st_handle,
	       unsigned char *buf, int n)
{
  /* add the trailer */
  int total_len = 48* ((n+8+CELL_DATA-1)/48);
  unsigned int crc, pti = 0;
  int i;

  printf("aal5_write: n=%d => total_len=%d\n",n,total_len);
  memset(buf+n,0,total_len-n);

  buf[total_len -6] = n>>8;
  buf[total_len -5] = n&0xff;

  crc = ~calc_crc(buf,total_len-4,-1);

  buf[total_len -4] = crc >> 24;
  buf[total_len -3] = crc >> 16;
  buf[total_len -2] = crc >>  8;
  buf[total_len -1] = crc;

  printf("AAL5:");
  for (i=0;i<total_len;i++)
    printf(" %02x",buf[i]);
  printf("\n");
  fflush(stdout);

  while (total_len>0)
    {
      if (total_len <= CELL_DATA)
	/* this is the last cell, so mark it */
	pti = 2;

      if (cell_write(st_handle,buf,CELL_DATA,MY_VPI,MY_VCI,pti) <0)
	return -1;

      buf       += CELL_DATA;
      total_len -= CELL_DATA;
    }

  return n;
}
 

int main(int argc, char *argv[])
{
  unsigned char buf[64*1024];
  int n, r;
  struct usb_dev_handle * st_handle;
  const char logfile[] = "pppoa.log";
  int fdin, fdout, log;
  pid_t pid;

  if (argc == 2)
    alternate = atoi(argv[1]);

  fdin = dup(0);
  fdout = dup(1);

  log = open(logfile,O_CREAT|O_RDWR|O_TRUNC,0666);
  if (log < 0)
    {
      perror(logfile);
      return -1;
    }

  close(0);
  close(1);
  close(2);

  dup(log);
  dup(log);
  dup(log);

  printf("fdin=%d, fdout=%d, log=%d\n",fdin,fdout,log);
  close(log);

  st_handle = usb_modem_init();
  if (!st_handle)
    {
      printf("unable to open USB modem!\n");
      return -1;
    }

  if ((pid=fork())!=0)
    {
      /* in parent process */

      for (;;)
	{
	  n = ppp_read(fdin,buf,sizeof(buf));
	  if (n<=0)
	    {
	      printf("ppp_read failed! Exiting ...\n");
	      fflush(stdout);
	      break;
	    }
	  r = aal5_write(st_handle,buf,n);
	}

      kill(getppid(),SIGTERM);
    }
  else
    {
      /* in child process */
      for (;;)
	{
	  n = aal5_read(st_handle,buf,sizeof(buf));
	  if (n<=0)
	    {
	      printf("aal5_read failed! Exiting ...\n");
	      fflush(stdout);
	      break;
	    }
	  r = ppp_write(fdout,buf,n);
	}
      
      kill(pid,SIGTERM);
    }

  return 0;
}
      
      
int test_main()
{
    /*
      Cell 1:
      00 80 02 30 ec c0 21 01 00 00 28 05 06 01 0d 17 	
      6e 0d 03 06 11 04 06 4e 13 17 01 01 17 c9 a0 59 	
      ab 4d ac a3 e4 77 21 ad 44 40 44 00 00 00 00 00 	
      00 00 00 00 00 
      
      Cell 2:
      00 80 02 32 ec 00 00 00 00 00 00 	
      00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 	
      00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 	
      00 00 00 00 00 2a ||8b d1 b8 dd||
      checksum
    */

  unsigned char pkt[1000] = {
    0xc0, 0x21, 0x01, 0x00, 0x00, 0x28, 0x05, 0x06,
    0x01, 0x0d, 0x17, 0x6e, 0x0d, 0x03, 0x06, 0x11,
    
    0x04, 0x06, 0x4e, 0x13, 0x17, 0x01, 0x01, 0x17,
    0xc9, 0xa0, 0x59, 0xab, 0x4d, 0xac, 0xa3, 0xe4,
    
    0x77, 0x21, 0xad, 0x44, 0x40, 0x44, 0x00, 0x00,
    0x00, 0x00
  };
  
  aal5_write(NULL,pkt,0x2a);
  return 0;
}

