/*
	Author   : Benoit PAPILLAULT
	Creation : 06/03/2001

07/03/2001: it works with 'modprobe usb-uhci' but not with 'modprobe uhci'.
	In this latter case, the error EPIPE is returned from the first
	ioctl.
*/
	


#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/ioctl.h>

#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

/* Copy/paste from kernel source, since it's protected by #idef __KERNEL__ */

/* Device descriptor */
struct usb_device_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;
        __u16 bcdUSB;
        __u8  bDeviceClass;
        __u8  bDeviceSubClass;
        __u8  bDeviceProtocol;
        __u8  bMaxPacketSize0;
        __u16 idVendor;
        __u16 idProduct;
        __u16 bcdDevice;
        __u8  iManufacturer;
        __u8  iProduct;
        __u8  iSerialNumber;
        __u8  bNumConfigurations;
} __attribute__ ((packed));

/*
	Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
	Creation : 05/03/2001

	Use usbdevfs directly (without libusb, which seems to be a crap).
	Take a look at /usr/src/linux/drivers/usb/devio.c for further
	information on IOCTL handle by USB device filesystem.
*/

/* Alcatel ADSL SpeedTouch USB modem idVendor/idProduct */

#define ST_VENDOR  0x06b9
#define ST_PRODUCT 0x4061

/* st_init : try to reproduce URB sent by the windows driver
   for this modem, URB got using Snoopy
*/

void st_init(int fd)
{
  struct usbdevfs_ctrltransfer ctrl;
  struct usb_device_descriptor desc;
  int r;

  printf("OK! I found the Alcatel ADSL SpeedTouch USB modem.\n");

  /* URB 1 : GET_DESCRIPTOR_FROM_DEVICE */

  ctrl.requesttype = 0x80;
  ctrl.request     = 6;
  ctrl.value       = 1;
  ctrl.index       = 0;
  ctrl.length      = sizeof(desc);
  ctrl.timeout     = 100; // EXPERIMENTAL
  ctrl.data        = &desc;

  memset(&desc,0,sizeof(desc));

  //  desc.bLength     = sizeof(desc);

  if ((r=ioctl(fd,USBDEVFS_CONTROL,&ctrl)) < 0)
    perror("USBDEVFS_CONTROL");

  printf("r=%d\n",r);
  printf("desc.bLength=%d\n",desc.bLength);

  /* we always get error EPIPE */
}

/* try to open the file and get USB device information,
   if we've found the modem, call st_init()
*/

void test_file(const char *path)
{
	int fd;
	struct usb_device_descriptor desc;

	fd = open(path,O_RDWR);
	if (fd == -1)
	{
		perror(path);
		return ;
	}

	if (read(fd,&desc,sizeof(desc)) == sizeof(desc))
	{
		/* great, we read something */

	  /* check, it match the correct structure */

	  if (desc.bLength == sizeof(desc))
	    {

		printf("=== %s ===\n",path);
		printf("  bLength         = %u\n",desc.bLength);
		printf("  bDescriptorType = %u\n",desc.bDescriptorType);
		printf("  bcdUSB          = %04x\n",desc.bcdUSB);
		printf("  idVendor        = %04x\n",desc.idVendor);
		printf("  idProduct       = %04x\n",desc.idProduct);
		printf("  bcdDevice       = %04x\n",desc.bcdDevice);

		if (desc.idVendor == ST_VENDOR
		    && desc.idProduct == ST_PRODUCT)
		  {
		    st_init(fd);
		  }
	    }
	}

	close(fd);
}

void scan_dir(const char *path)
{
	DIR * dir;
	struct dirent * dirp;

	dir = opendir(path);
	if (dir == NULL)
	{
		perror(path);
		return ;
	}

	while ((dirp=readdir(dir)) != NULL)
	{
		char file[PATH_MAX+1];
		struct stat statbuf;

		if (strcmp(dirp->d_name,".")==0)
			continue;
		if (strcmp(dirp->d_name,"..")==0)
			continue;

		sprintf(file,"%s/%s",path,dirp->d_name);

		if (stat(file,&statbuf)!=0)
		{
			perror(file);
			continue;
		}

		if (S_ISDIR(statbuf.st_mode))
		{
			scan_dir(file);
			continue;
		}

		if (S_ISREG(statbuf.st_mode))
			test_file(file);
	}

	closedir(dir);
}


int main()
{
	scan_dir("/proc/bus/usb");
	return 0;
}

